#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#define ASLFLOAT double
#define ASLINT int
#define ASLINT_INF INT_MAX
#ifndef NULL
#define NULL 0
#endif

/* ========================================================================== */
/* === structures =========================================================== */
/* ========================================================================== */

typedef struct ASLobjective_struct /* passed to user's evaluation routines */
{
    /* if objective function not quadratic: */
    ASLFLOAT     f ; /* value at x (provided by user) */
    ASLFLOAT  fold ; /* final value of f at prior iteration (from ASL) */
    ASLFLOAT    *x ; /* current iterate (from ASL) */
    ASLFLOAT *xold ; /* final x at prior iteration (from ASL) */
    ASLFLOAT    *g ; /* gradient at x (provided by user) */
    ASLFLOAT *gold ; /* previous gradient (from ASL) */
    ASLFLOAT    *d ; /* nonzero components of search direction (from ASL) */
    int         *i ; /* indices of nonzeros elements */
    int         nz ; /* number of nonzeros in d */
    ASLFLOAT  step ; /* evaluate gradient at x = xold + (*step)*d (from ASL) */
    ASLFLOAT deriv1; /* when ASL computes derivative in search direction at
                        start of iteration (info = 0), it stores it in deriv1 */
    int       info ; /* -1 = initial request for value/grad
                         0 = first evaluation in search direction
                         1 = later evaluation in search direction (from ASL) */
    int          n ; /* problem dimension (from ASL) */

    /* If objective function is quadratic (Parm->QuadCost = TRUE),
       then f(x) = .5x'Ax + c'x. The product A*d should be return in p
       during each call to the user Ad routine */
    ASLFLOAT    *p ; /* the product A*d (provided by user)*/
} ASLobjective ;

typedef struct ASLparm_struc /* user parameters (default in ASLdefault) */
{
    ASLFLOAT    epsmu0 ; /* error tol for mu: B - A <= epsmu0*|g|/|a| */
    ASLFLOAT    epsmu1 ; /* error tol for mu: B - A <= epsmu1*(|A|+|B|) */
    ASLFLOAT   lambda0 ; /* lower bound for lambda_k */
    ASLFLOAT lambda0Factor ;/* require that lambda_k >= lambda0Factor*||g|| */
    ASLFLOAT Armijo_delta ; /* f(st) <= fR+st*delta*f'(0), fR = mfax or fr */
    ASLFLOAT Armijo_decay ; /* backtracking factor in Armijo line search */
    int  max_backsteps ; /* max number of Armijo backsteps */
    ASLFLOAT     gamma ; /* required decay factor of width in bracket interval*/
    int     max_mu_its ; /* max number of iterations when computing mu */
    ASLINT      maxits ; /* maximum number of iterations */
    int              M ; /* memory for nonmonotone line search */
    int     PrintParms ; /* TRUE means print parameter values */
    int     PrintLevel ; /* Level 0  = no print, ... , Level 2 = maximum print*/
    int     PrintFinal ; /* T => print final statistics, F => no */
    /* T => use approximate nonmonotone Armijo line search
       F => use ordinary nonmonotone Armijo line search, switch to
            approximate Armijo when |f_r-f| < AArmijoFac*|min (f_r, f_{max})| */
    int        AArmijo ;
    ASLFLOAT AArmijoFac ;
    /* T => estimated error in function value = eps*|min (f_r, f_{max}) |
       F => estimated error in function value = eps */
    int       PertRule ;
    ASLFLOAT       eps ;
    /* search for non nan function value by shrinking search interval
       at most nshrink times */
    int       nshrink ;
    /* factor by which interval shrinks when searching for non nan value */
    ASLFLOAT  nan_fac ;
    int             L ; /* update fr if fmin was not improved after
                                L iterations */
    int             P ; /* update fr if P previous initial stepsize was
                                all accepted */
    int            nm ; /* nominal CBB cycle length */
    int      MaxCycle ; /* absolute max number of iterations in a cycle */
    ASLFLOAT   gamma1 ; /* criterion for updating reference value fr */
    ASLFLOAT   gamma2 ; /* criterion for updating reference value fr */
    ASLFLOAT  armijo0 ; /* criterion for quadratic interpolation in
                                cbb line search */
    ASLFLOAT  armijo1 ; /* criterion for quadratic interpolation in
                                cbb line search */
    ASLFLOAT   lo_cut ; /* g_j > 0 and x_j - lo_j < lo_cut => d_j = 0 */
    ASLFLOAT   hi_cut ; /* g_j < 0 and hi_j - x_j < hi_cut => d_j = 0 */
    ASLFLOAT   big_yi ; /* neglect components of y which satisfy
                           yi >= y_s*big_yi, y_s = y_i */
    int        a_is_1 ; /* = 1 if the a vector is identically 1 (0 otherwise) */
    int      QuadCost ; /* = 1 means the cost is quadratic (0 otherwise) */
    int      QuadLazy ; /* objective function is quadratic, let code compute
                           starting gradient Ax + c and starting cost
                           .5x'Ax + c'x */
    int bnd_is_scalar ; /* 1 means scalar bounds (0 otherwise) */
    int      Subspace ; /* TRUE means use subspace implementation */
    int          nsub ; /* dimension of subspace */
    ASLFLOAT     fsub ; /* if nsub < 0, dimension = fsub*n */
} ASLparm ;

typedef struct ASLstat_struct /* statistics returned to user */
{
    ASLFLOAT     f ; /* function value */
    ASLFLOAT   err ; /* max norm of g - mu d */
    ASLINT      it ; /* number of iterations */
    ASLINT      nf ; /* number of function evaluations */
    ASLINT      ng ; /* number of gradient evaluations*/
    ASLINT     nmu ; /* total number of iterations in computation of mu */
} ASLstat ;

/* prototypes */

int ASL /* return:
                   0 (error tolerance satisfied)
                   1 (number of iterations exceeded limit)
                   2 (out of memory)
                   3 (line search failed)
                   4 (search direction not a descent direction)
                   5 (computation of search direction fails)
                   6 (search direction not a descent direction )
                   7 (computation of mu fails)
                   8 (function value is nan in line search)*/
(
    ASLFLOAT      *x, /* input: starting guess, output: the solution */
    ASLFLOAT     *lo, /* lower bounds (single bound if bndType = 1) */
    ASLFLOAT     *hi, /* upper bounds (single bound if bndType = 1) */
    ASLFLOAT      *a, /* linear constraint vector in a'x = b
                         NULL = no linear constraint or a identically 1
                         if Uparm->a_is_1 = 1 */
    ASLFLOAT       b, /* right side */
    ASLFLOAT     *g0, /* if cost quadratic, g0 = Ax + c = starting gradient */
    int            n, /* problem dimension */
    ASLFLOAT grad_tol,/* bound on KKT error */
    /* any of the remaining arguments can be NULL */
    ASLstat    *Stat, /* structure with statistics (can be NULL) */
    ASLparm   *Uparm, /* user parameters, NULL = use default parameters */
    void      (*Ad) (ASLobjective *), /* compute A*d if quadratic objective */
    void   (*value) (ASLobjective *), /* evaluate objective function */
    void    (*grad) (ASLobjective *), /* evaluate objective gradient */
    void (*valgrad) (ASLobjective *)  /* evaluate function and gradient
                                         NULL = use value & grad routines */
) ;

void ASLdefault
(
    ASLparm *Parm /* pointer to parameter structure */
) ;
